/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "client/src/cfg.hh"
#include "client/src/generate_zone.hh"
#include "client/src/get_zone_data.hh"
#include "client/src/get_zones_to_generate.hh"
#include "client/src/hook_exec.hh"

#include "liblog/liblog.hh"
#include "liblog/log.hh"

#include "fred_api/zone/service_zone_grpc.grpc.pb.h"

#include <boost/asio/ip/address.hpp>
#include <boost/filesystem.hpp>
#include <boost/optional.hpp>
#include <boost/program_options.hpp>

#include <grpc++/grpc++.h>
#include <grpc/support/log.h>

#include <algorithm>
#include <cstdint>
#include <fstream>
#include <functional>
#include <future>
#include <iostream>
#include <stdexcept>
#include <string>
#include <thread>
#include <tuple>
#include <utility>

namespace Fred {
namespace Zone {
namespace Client {

namespace {

void set_logging(const Cfg::Options::Log& log_options);

Cfg::Options::Zone get_zone_options(
        const std::string& zone,
        const Cfg::Options::ParticularZones& zones_options,
        const Cfg::Options::Zone& fallback_options);

std::uint32_t process_zone_generation(
        const std::string& zone,
        const Cfg::Options::Zone& zone_options,
        const Cfg::Options::ZoneService& zone_service_options,
        std::map<std::string, std::ofstream>& bind_config_files);

void view(const std::string& zone, const Cfg::Options::Zone& zone_config);

struct TooDifferent
{
    std::uint64_t absolute;
    double relative;
};

}//namespace Fred::Zone::Client::{anonymous}

}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred

int main(int argc, char* argv[])
{
    Fred::Zone::Client::Cfg::handle_cli_args(argc, argv);
    try
    {
        const auto& cfg = Fred::Zone::Client::Cfg::Options::get();
        Fred::Zone::Client::set_logging(cfg.log);
        LIBLOG_SET_CONTEXT(Ctx, log_ctx, __func__);
        const auto zones_to_generate = Fred::Zone::Client::get_zones_to_generate(cfg.generate, cfg.zone_service);
        std::map<std::string, std::ofstream> bind_config_files{};
        std::for_each(begin(zones_to_generate), end(zones_to_generate), [&](const std::string& zone)
        {
            const auto zone_options = Fred::Zone::Client::get_zone_options(zone, cfg.zones, cfg.generic.zone_common);
            Fred::Zone::Client::view(zone, zone_options);
            try
            {
                const auto zone_soa_id = Fred::Zone::Client::process_zone_generation(zone, zone_options, cfg.zone_service, bind_config_files);
                LIBLOG_INFO("zone file with id {} successfully generated", zone_soa_id);
            }
            catch (const Fred::Zone::Client::TooDifferent&) { }
        });
        std::for_each(begin(bind_config_files), end(bind_config_files), [](auto& bind_config_file)
        {
            bind_config_file.second.close();
            LIBLOG_INFO("bind config file {} successfully generated", bind_config_file.first);
        });
        return EXIT_SUCCESS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }
}

namespace Fred {
namespace Zone {
namespace Client {

namespace {

Cfg::Options::Zone get_zone_options(
        const std::string& zone,
        const Cfg::Options::ParticularZones& zones_options,
        const Cfg::Options::Zone& fallback_options)
{
    const auto zone_options_iter = zones_options.find(zone);
    const bool has_zone_options = zone_options_iter != zones_options.end();
    auto zone_options = has_zone_options ? zone_options_iter->second
                                         : Cfg::Options::Zone{};
    if (zone_options.zone_directory == boost::none)
    {
        zone_options.zone_directory = fallback_options.zone_directory;
        if (zone_options.zone_directory == boost::none)
        {
            zone_options.zone_directory = ".";
        }
    }
    zone_options.no_backups |= fallback_options.no_backups;
    if (!zone_options.no_backups)
    {
        if (zone_options.backup_directory == boost::none)
        {
            zone_options.backup_directory = fallback_options.backup_directory;
            if (zone_options.backup_directory == boost::none)
            {
                zone_options.backup_directory = zone_options.zone_directory;
            }
        }
    }
    else
    {
        zone_options.backup_directory = boost::none;
    }
    if ((zone_options.max_changes_absolute == boost::none) && (zone_options.max_changes_relative == boost::none))
    {
        zone_options.max_changes_absolute = fallback_options.max_changes_absolute;
        zone_options.max_changes_relative = fallback_options.max_changes_relative;
    }
    if (zone_options.bind_config_file_name == boost::none)
    {
        zone_options.bind_config_file_name = fallback_options.bind_config_file_name;
    }
    if (zone_options.header_file_name == boost::none)
    {
        zone_options.header_file_name = fallback_options.header_file_name;
    }
    if (zone_options.footer_file_name == boost::none)
    {
        zone_options.footer_file_name = fallback_options.footer_file_name;
    }
    if (zone_options.post_hook_command == boost::none)
    {
        zone_options.post_hook_command = fallback_options.post_hook_command;
        zone_options.post_hook_arguments = fallback_options.post_hook_arguments;
    }
    return zone_options;
}

void view(const std::string& zone, const Cfg::Options::Zone& zone_config)
{
    std::cout << "generate zone file for " << zone << " zone" << std::endl;
    std::cout << "  zone_directory: " << *zone_config.zone_directory << std::endl;
    if (!zone_config.no_backups)
    {
        std::cout << "  backup_directory: " << *zone_config.backup_directory << std::endl;
        if (zone_config.max_changes_absolute != boost::none)
        {
            std::cout << "  max_changes_absolute: " << *zone_config.max_changes_absolute << std::endl;
        }
        if (zone_config.max_changes_relative != boost::none)
        {
            std::cout << "  max_changes_relative: " << *zone_config.max_changes_relative << std::endl;
        }
    }
    if (zone_config.bind_config_file_name != boost::none)
    {
        std::cout << "  bind_config_file_name: " << *zone_config.bind_config_file_name << std::endl;
    }
    if (zone_config.header_file_name != boost::none)
    {
        std::cout << "  header_file_name: " << *zone_config.header_file_name << std::endl;
    }
    if (zone_config.footer_file_name != boost::none)
    {
        std::cout << "  footer_file_name: " << *zone_config.footer_file_name << std::endl;
    }
    if (zone_config.post_hook_command != boost::none)
    {
        std::cout << "  post_hook_command: " << *zone_config.post_hook_command << std::endl;
        std::for_each(begin(zone_config.post_hook_arguments), end(zone_config.post_hook_arguments), [](const std::string& argument)
        {
            std::cout << "    post_hook_argument: " << argument << std::endl;
        });
    }
}

template <typename T = void>
struct DataFromThread
{
    DataFromThread()
        : promise{},
          future{promise.get_future()}
    { }
    std::promise<T> promise;
    std::future<T> future;
};

void run_get_zone_data(const std::string& zone_file, std::promise<ZoneData> result)
{
    try
    {
        result.set_value(get_zone_data(zone_file));
    }
    catch (...)
    {
        result.set_exception(std::current_exception());
    }
}

std::uint32_t generate_zone_and_check_similarity(
        const std::string& zone,
        const Cfg::Options::Zone& zone_options,
        const Cfg::Options::ZoneService& zone_service_options,
        const std::string& old_zone_file_name,
        std::ostream& new_zone_file)
{
    DataFromThread<ZoneData> old_zone_data_from_thread;
    ZoneData new_zone_data;
    auto thread = std::thread{
            run_get_zone_data,
            std::cref(old_zone_file_name),
            std::move(old_zone_data_from_thread.promise)};
    bool is_joined = false;
    try
    {
        const auto zone_soa_id = generate_zone(zone, zone_service_options, new_zone_file, &new_zone_data);
        const auto old_zone_data = old_zone_data_from_thread.future.get();
        thread.join();
        is_joined = true;
        const auto difference = diff(old_zone_data, new_zone_data);
        const bool are_similar = zone_options.max_changes_absolute != boost::none
                ? difference.absolute_change() <= *zone_options.max_changes_absolute
                : difference.change_ratio() <= *zone_options.max_changes_relative;
        if (are_similar)
        {
            LIBLOG_INFO("difference is {:0.6f}%", difference.change_ratio() * 100);
            return zone_soa_id;
        }
        throw TooDifferent{difference.absolute_change(), difference.change_ratio()};
    }
    catch (const TooDifferent& e)
    {
        LIBLOG_ERROR("difference is {:0.6f}%", e.relative * 100);
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("exception caught in zone generator: {}", e.what());
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown exception in zone generator");
    }
    if (!is_joined)
    {
        thread.join();
    }
    throw std::runtime_error{"unable to generate zone file"};
}

void add_bind_configuration(
        const std::string& bind_config_file_name,
        const std::string& zone,
        const std::string& zone_file_name,
        std::map<std::string, std::ofstream>& bind_config_files)
{
    auto bind_config_file_iter = bind_config_files.find(bind_config_file_name);
    if (bind_config_file_iter == bind_config_files.end())
    {
        bind_config_file_iter = bind_config_files.emplace(bind_config_file_name, std::ofstream{bind_config_file_name}).first;
    }
    else
    {
        bind_config_file_iter->second << "\n";
    }
    bind_config_file_iter->second << "zone \"" << zone << "\" IN {\n"
                                     "     type master;\n"
                                     "     file \"" << zone_file_name << "\";\n"
                                     "};\n";
}

std::uint32_t process_zone_generation(
        const std::string& zone,
        const Cfg::Options::Zone& zone_options,
        const Cfg::Options::ZoneService& zone_service_options,
        std::map<std::string, std::ofstream>& bind_config_files)
{
    const auto tmp_zone_file_name = boost::filesystem::path{*zone_options.zone_directory}.append("<db." + zone + ".new>");
    const auto zone_file_name = boost::filesystem::path{*zone_options.zone_directory}.append("db." + zone);
    const bool zone_file_exists = is_regular_file(zone_file_name);
    const bool check_difference =
            zone_file_exists &&
            ((zone_options.max_changes_absolute != boost::none) ||
             (zone_options.max_changes_relative != boost::none));
    std::cout << "  zone file: " << tmp_zone_file_name << std::endl;
    static const auto append = [](std::ostream& dst, const boost::optional<std::string>& src_file)
    {
        if (src_file != boost::none)
        {
            auto src = std::ifstream{*src_file};
            std::string line;
            while (getline(src, line))
            {
                dst << line << "\n";
            }
        }
    };
    auto tmp_zone_file = std::ofstream{tmp_zone_file_name.string()};
    append(tmp_zone_file, zone_options.header_file_name);
    const auto zone_soa_id = check_difference ? generate_zone_and_check_similarity(zone, zone_options, zone_service_options, zone_file_name.string(), tmp_zone_file)
                                              : generate_zone(zone, zone_service_options, tmp_zone_file);
    append(tmp_zone_file, zone_options.footer_file_name);
    tmp_zone_file.close();
    static const auto uniform = [](const boost::filesystem::path& path)
    {
        return canonical(absolute(path));
    };
    const auto make_backup_zone_file_name = [&]()
    {
        return boost::filesystem::path{*zone_options.backup_directory}.append("db." + zone + "-" + std::to_string(zone_soa_id));
    };
    if (!zone_options.no_backups && zone_file_exists)
    {
        const auto backup_zone_file_name = make_backup_zone_file_name();
        rename(zone_file_name, backup_zone_file_name);
    }
    rename(tmp_zone_file_name, zone_file_name);
    if (zone_options.bind_config_file_name != boost::none)
    {
        add_bind_configuration(
                    uniform(*zone_options.bind_config_file_name).string(),
                    zone,
                    uniform(zone_file_name).string(),
                    bind_config_files);
    }
    if (zone_options.post_hook_command != boost::none)
    {
        const auto backup = make_backup_zone_file_name();
        if (is_regular_file(backup))
        {
            hook_exec(zone, uniform(zone_file_name).string(), uniform(backup).string(), zone_options);
        }
        else
        {
            hook_exec(zone, uniform(zone_file_name).string(), "", zone_options);
        }
    }
    return zone_soa_id;
}

void disable_grpc_library_logging()
{
    static const auto dummy_log_func = [](::gpr_log_func_args*) { };
    ::gpr_set_log_function(dummy_log_func);
    static constexpr auto do_not_log = static_cast<::gpr_log_severity>(GPR_LOG_SEVERITY_ERROR + 1);
    ::gpr_set_log_verbosity(do_not_log);
}

void grpc_log(::gpr_log_func_args* log_args)
{
    try
    {
        switch (log_args->severity)
        {
            case GPR_LOG_SEVERITY_DEBUG:
                LIBLOG_DEBUG("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
            case GPR_LOG_SEVERITY_INFO:
                LIBLOG_INFO("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
            case GPR_LOG_SEVERITY_ERROR:
                LIBLOG_ERROR("{} at {}:{}", log_args->message, log_args->file, log_args->line);
                return;
        }
        LIBLOG_WARNING("{} at {}:{}", log_args->message, log_args->file, log_args->line);
    }
    catch (...)
    { }
}

constexpr auto no_log_grpc_severity = static_cast<::gpr_log_severity>(GPR_LOG_SEVERITY_ERROR + 1);

constexpr ::gpr_log_severity make_grpc_log_severity(LibLog::Level min_severity)
{
    switch (min_severity)
    {
        case LibLog::Level::critical:
            return no_log_grpc_severity;
        case LibLog::Level::error:
            return GPR_LOG_SEVERITY_ERROR;
        case LibLog::Level::warning:
        case LibLog::Level::info:
            return GPR_LOG_SEVERITY_INFO;
        case LibLog::Level::debug:
        case LibLog::Level::trace:
            return GPR_LOG_SEVERITY_DEBUG;
    }
    return no_log_grpc_severity;
}

decltype(auto) get_min_severity(const Cfg::Options::Log& log_options)
{
    boost::optional<LibLog::Level> min_severity{false, LibLog::Level::critical};
    const auto update_min_severity = [&](LibLog::Level severity)
    {
        if ((min_severity == boost::none) ||
            (severity < *min_severity))
        {
            min_severity = severity;
        }
    };
    if (log_options.console != boost::none)
    {
        update_min_severity(log_options.console->min_severity);
    }
    if (log_options.file != boost::none)
    {
        update_min_severity(log_options.file->min_severity);
    }
    return min_severity;
}

void set_grpc_library_logging(const Cfg::Options::Log& log_options)
{
    if (!log_options.log_grpc_library_events)
    {
        LIBLOG_INFO("do not log gRPC library");
        disable_grpc_library_logging();
        return;
    }
    const auto min_severity = get_min_severity(log_options);
    if (min_severity == boost::none)
    {
        disable_grpc_library_logging();
        return;
    }
    ::gpr_set_log_function(grpc_log);
    ::gpr_set_log_verbosity(make_grpc_log_severity(*min_severity));
    LIBLOG_INFO("log gRPC library");
}

decltype(auto) make_sink_config(const Cfg::Options::Log::Console& log_options)
{
    LibLog::Sink::ConsoleSinkConfig sink_config;
    sink_config.set_level(log_options.min_severity);
    sink_config.set_output_stream(log_options.output_stream);
    sink_config.set_color_mode(log_options.color_mode);
    return sink_config;
}

decltype(auto) make_sink_config(const Cfg::Options::Log::File& log_options)
{
    LibLog::Sink::FileSinkConfig sink_config{log_options.file_name};
    sink_config.set_level(log_options.min_severity);
    return sink_config;
}

decltype(auto) make_sink_config(const Cfg::Options::Log::Syslog& log_options)
{
    LibLog::Sink::SyslogSinkConfig sink_config{};
    sink_config.set_level(log_options.min_severity);
    sink_config.set_syslog_facility(log_options.facility);
    sink_config.set_syslog_ident(log_options.ident);
    sink_config.set_syslog_options(log_options.options);
    return sink_config;
}

void set_logging(const Cfg::Options::Log& log_options)
{
    bool has_device = false;
    LibLog::LogConfig log_config;
    if (log_options.console != boost::none)
    {
        try
        {
            log_config.add_sink_config(make_sink_config(*log_options.console));
            has_device = true;
        }
        catch (...) { }
    }
    if (log_options.file != boost::none)
    {
        try
        {
            log_config.add_sink_config(make_sink_config(*log_options.file));
            has_device = true;
        }
        catch (...) { }
    }
    if (log_options.syslog != boost::none)
    {
        try
        {
            log_config.add_sink_config(make_sink_config(*log_options.syslog));
            has_device = true;
        }
        catch (...) { }
    }
    if (has_device)
    {
        try
        {
            LibLog::Log::start<LibLog::ThreadMode::multi_threaded>(log_config);
            set_grpc_library_logging(log_options);
            return;
        }
        catch (const std::exception& e)
        {
            std::cerr << "Setting up logging failure: " << e.what() << std::endl;
            ::exit(EXIT_FAILURE);
        }
        catch (...)
        {
            std::cerr << "Setting up logging failure: Unexpected exception caught" << std::endl;
            ::exit(EXIT_FAILURE);
        }
    }
    try
    {
        disable_grpc_library_logging();
    }
    catch (...)
    {
        std::cerr << "unable to disable gRPC library logging" << std::endl;
    }
    std::cout << "no logging is configured!?" << std::endl;
}

}//namespace Fred::Zone::Client::{anonymous}

}//namespace Fred::Zone::Client
}//namespace Fred::Zone
}//namespace Fred
